"""Database model for Jobs."""

from typing import TYPE_CHECKING, List

from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from lm_api.api.models.crud_base import CrudBase

if TYPE_CHECKING:
    from lm_api.api.models.booking import Booking
else:
    Booking = "Booking"


class Job(CrudBase):
    """
    Represents the jobs submitted in a cluster.
    """

    slurm_job_id = mapped_column(String, nullable=False)
    cluster_client_id = mapped_column(String, nullable=False)
    username = mapped_column(String, nullable=False)
    lead_host = mapped_column(String, nullable=False)

    bookings: Mapped[List[Booking]] = relationship(
        Booking,
        back_populates="job",
        lazy="selectin",
        cascade="all, delete-orphan",
    )

    searchable_fields = [slurm_job_id, username, lead_host]
    sortable_fields = [slurm_job_id, cluster_client_id, username, lead_host]

    def __repr__(self):
        return (
            f"Job(id={self.id}, "
            f"slurm_job_id={self.slurm_job_id}, "
            f"cluster_client_id={self.cluster_client_id}, "
            f"username={self.username}, "
            f"lead_host={self.lead_host})"
        )



from typing import List, Optional

from fastapi import APIRouter, Body, Depends, Query, status

from lm_api.api.cruds.booking import BookingCRUD
from lm_api.api.models.booking import Booking
from lm_api.api.schemas.booking import BookingCreateSchema, BookingSchema
from lm_api.database import SecureSession, secure_session
from lm_api.permissions import Permissions

router = APIRouter()


crud_booking = BookingCRUD(Booking)


@router.post(
    "",
    response_model=BookingSchema,
    status_code=status.HTTP_201_CREATED,
)
async def create_booking(
    booking: BookingCreateSchema = Body(..., description="Booking to be created"),
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.BOOKING_CREATE)),
):
    """Create a new booking."""
    return await crud_booking.create(db_session=secure_session.session, obj=booking)


@router.get(
    "",
    response_model=List[BookingSchema],
    status_code=status.HTTP_200_OK,
)
async def read_all_bookings(
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.BOOKING_READ, commit=False)
    ),
):
    """Return all bookings."""
    return await crud_booking.read_all(
        db_session=secure_session.session,
        sort_field=sort_field,
        sort_ascending=sort_ascending,
    )


@router.get(
    "/{booking_id}",
    response_model=BookingSchema,
    status_code=status.HTTP_200_OK,
)
async def read_booking(
    booking_id: int,
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.BOOKING_READ, commit=False)
    ),
):
    """Return a booking with associated bookings with the given id."""
    return await crud_booking.read(db_session=secure_session.session, id=booking_id)


@router.delete(
    "/{booking_id}",
    status_code=status.HTTP_200_OK,
)
async def delete_booking(
    booking_id: int,
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.BOOKING_DELETE)),
):
    """Delete a booking from the database."""
    return await crud_booking.delete(db_session=secure_session.session, id=booking_id)



"""Database model for Configurations."""

from typing import TYPE_CHECKING, List

from sqlalchemy import Integer, String
from sqlalchemy.orm import Mapped, declared_attr, mapped_column, relationship
from sqlalchemy.sql.schema import CheckConstraint

from lm_api.api.models.crud_base import CrudBase

if TYPE_CHECKING:
    from lm_api.api.models.feature import Feature
    from lm_api.api.models.license_server import LicenseServer
else:
    Feature = "Feature"
    LicenseServer = "LicenseServer"


class Configuration(CrudBase):
    """
    Represents the feature configurations.
    """

    @declared_attr.directive
    def __tablename__(cls) -> str:
        return "configs"

    name = mapped_column(String, nullable=False)
    cluster_client_id = mapped_column(String, nullable=False)
    grace_time = mapped_column(Integer, CheckConstraint("grace_time>=0"), nullable=False)
    type = mapped_column(String, nullable=False)

    license_servers: Mapped[List[LicenseServer]] = relationship(
        LicenseServer,
        back_populates="configurations",
        lazy="selectin",
        cascade="all, delete-orphan",
        uselist=True,
    )
    features: Mapped[List[Feature]] = relationship(
        Feature,
        back_populates="configurations",
        lazy="selectin",
        cascade="all, delete-orphan",
        uselist=True,
    )

    searchable_fields = [name]
    sortable_fields = [name, cluster_client_id, type]

    def __repr__(self):
        return (
            f"Config(id={self.id}, "
            f"name={self.name}, "
            f"cluster_client_id={self.cluster_client_id}, "
            f"grace_time={self.grace_time}), "
            f"type={self.type}), "
        )



"""Feature schemas for the License Manager API."""
from typing import Optional

from pydantic import ConfigDict, BaseModel, Field, NonNegativeInt

from lm_api.api.schemas.base import BaseCreateSchema, BaseUpdateSchema
from lm_api.api.schemas.product import ProductSchema


class FeatureCreateSchema(BaseCreateSchema):
    """
    Represents the features in a feature configuration.
    """

    name: str = Field(
        ..., title="Name of the feature", max_length=255, description="The name of the feature."
    )
    product_id: int = Field(..., title="Product ID", description="The ID of the product of the feature.")
    config_id: int = Field(
        ..., title="Configuration ID", description="The ID of the configuration that the feature belongs to."
    )
    reserved: int = Field(
        0,
        ge=0,
        le=2**31 - 1,
        title="Reserved quantity",
        description="The quantity of the feature that is reserved for usage in desktop environments.",
    )


class FeatureWithoutConfigIdCreateSchema(BaseCreateSchema):
    """
    Represents the features in a feature configuration.
    """

    name: str = Field(
        ..., title="Name of the feature", max_length=255, description="The name of the feature."
    )
    product_id: int = Field(..., title="Product ID", description="The ID of the product of the feature.")
    reserved: int = Field(
        0,
        ge=0,
        le=2**31 - 1,
        title="Reserved quantity",
        description="The quantity of the feature that is reserved for usage in desktop environments.",
    )


class FeatureWithOptionalIdUpdateSchema(BaseUpdateSchema):
    """
    Feature to be updated in the database.
    """

    id: Optional[int] = Field(None, title="ID", description="The ID of the feature.")
    name: Optional[str] = Field(
        None, title="Name of the feature", max_length=255, description="The name of the feature."
    )
    product_id: Optional[int] = Field(
        None, title="Product ID", description="The ID of the product of the feature."
    )
    reserved: Optional[int] = Field(
        None,
        ge=0,
        le=2**31 - 1,
        title="Reserved quantity",
        description="The quantity of the feature that is reserved for usage in desktop environments.",
    )


class FeatureUpdateSchema(BaseUpdateSchema):
    """
    Represents the features in a feature configuration.
    """

    name: Optional[str] = Field(
        None, title="Name of the feature", max_length=255, description="The name of the feature."
    )
    product_id: Optional[int] = Field(
        None, title="Product ID", description="The ID of the product of the feature."
    )
    config_id: Optional[int] = Field(
        None, title="Configuration ID", description="The ID of the configuration that the feature belongs to."
    )
    reserved: Optional[int] = Field(
        None,
        ge=0,
        le=2**31 - 1,
        title="Reserved quantity",
        description="The quantity of the feature that is reserved for usage in desktop environments.",
    )
    total: Optional[int] = Field(
        None,
        ge=0,
        le=2**31 - 1,
        title="Total quantity",
        description="The total quantity of licenses.",
    )
    used: Optional[int] = Field(
        None,
        ge=0,
        le=2**31 - 1,
        title="Used quantity",
        description="The quantity of the feature that is used.",
    )


class FeatureUpdateByNameSchema(BaseUpdateSchema):
    """
    Represents the feature usage data that will be updated using the
    product and feature name as a filter.
    """

    product_name: str = Field(
        ..., title="Product name", max_length=255, description="The name of the product."
    )
    feature_name: str = Field(
        ..., title="Feature name", max_length=255, description="The name of the feature."
    )
    total: int = Field(
        0,
        ge=0,
        le=2**31 - 1,
        title="Total quantity",
        description="The total quantity of licenses.",
    )
    used: int = Field(
        0,
        ge=0,
        le=2**31 - 1,
        title="Used quantity",
        description="The quantity of the feature that is used.",
    )


class FeatureSchema(BaseModel):
    """
    Represents the features in a feature configuration.
    """

    id: int = Field(..., title="ID", description="The ID of the feature.")
    name: str = Field(
        ..., title="Name of the feature", max_length=255, description="The name of the feature."
    )
    product: ProductSchema = Field(
        ..., title="Product of the feature", description="The product of the feature."
    )
    config_id: int = Field(
        ..., title="Configuration ID", description="The ID of the configuration that the feature belongs to."
    )
    reserved: NonNegativeInt = Field(
        0,
        title="Reserved quantity",
        description="The quantity of the feature that is reserved for usage in desktop environments.",
    )
    total: NonNegativeInt = Field(
        0,
        title="Total quantity",
        description="The total quantity of licenses.",
    )
    used: NonNegativeInt = Field(
        0,
        title="Used quantity",
        description="The quantity of the feature that is used.",
    )
    booked_total: Optional[NonNegativeInt] = Field(
        0,
        title="Booked total quantity",
        description="The total quantity of licenses that are booked.",
    )
    model_config = ConfigDict(from_attributes=True)

    @classmethod
    def from_flat_dict(cls, d):
        return cls(
            product=ProductSchema(
                id=d["product_id"],
                name=d["product_name"],
            ),
            **{k: v for (k, v) in d.items() if not k.startswith("product")},
        )



from typing import List, Optional

from fastapi import APIRouter, Body, Depends, HTTPException, Query, status

from lm_api.api.cruds.configuration import ConfigurationCRUD
from lm_api.api.cruds.feature import FeatureCRUD
from lm_api.api.cruds.generic import GenericCRUD
from lm_api.api.models.configuration import Configuration
from lm_api.api.models.feature import Feature
from lm_api.api.models.license_server import LicenseServer
from lm_api.api.models.product import Product
from lm_api.api.schemas.configuration import (
    ConfigurationCompleteCreateSchema,
    ConfigurationCompleteUpdateSchema,
    ConfigurationCreateSchema,
    ConfigurationSchema,
    ConfigurationUpdateSchema,
)
from lm_api.api.schemas.feature import FeatureCreateSchema, FeatureUpdateSchema
from lm_api.api.schemas.license_server import LicenseServerCreateSchema, LicenseServerUpdateSchema
from lm_api.database import SecureSession, secure_session
from lm_api.permissions import Permissions

router = APIRouter()


crud_configuration = ConfigurationCRUD(Configuration)
crud_product = GenericCRUD(Product)
crud_feature = FeatureCRUD(Feature)
crud_license_server = GenericCRUD(LicenseServer)


@router.post(
    "",
    response_model=ConfigurationSchema,
    status_code=status.HTTP_201_CREATED,
)
async def create_configuration(
    configuration: ConfigurationCompleteCreateSchema = Body(..., description="Configuration to be created"),
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.CONFIG_CREATE)),
):
    """
    Create a new configuration.

    The features for the configuration will be specified in the request body.
    If the feature's product doesn't exist, it will be created.
    """
    configuration_created = await crud_configuration.create(
        db_session=secure_session.session,
        obj=ConfigurationCreateSchema(**configuration.dict(exclude={"features", "license_servers"})),
    )

    if configuration.features:
        try:
            for feature in configuration.features:
                feature_obj = {
                    "name": feature.name,
                    "product_id": feature.product_id,
                    "config_id": configuration_created.id,
                    "reserved": feature.reserved,
                }
                await crud_feature.create(
                    db_session=secure_session.session, obj=FeatureCreateSchema(**feature_obj)
                )
        except HTTPException:
            await crud_configuration.delete(db_session=secure_session.session, id=configuration_created.id)
            raise

    if configuration.license_servers:
        try:
            for license_server in configuration.license_servers:
                license_server_obj = {
                    "config_id": configuration_created.id,
                    "host": license_server.host,
                    "port": license_server.port,
                }
                await crud_license_server.create(
                    db_session=secure_session.session, obj=LicenseServerCreateSchema(**license_server_obj)
                )
        except HTTPException:
            await crud_configuration.delete(db_session=secure_session.session, id=configuration_created.id)
            raise

    return await crud_configuration.read(
        db_session=secure_session.session, id=configuration_created.id, force_refresh=True
    )


@router.get(
    "",
    response_model=List[ConfigurationSchema],
    status_code=status.HTTP_200_OK,
)
async def read_all_configurations(
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.CONFIG_READ, commit=False)
    ),
):
    """Return all configurations with the associated license servers and features."""
    return await crud_configuration.read_all(
        db_session=secure_session.session,
        search=search,
        sort_field=sort_field,
        sort_ascending=sort_ascending,
    )


@router.get(
    "/by_client_id",
    response_model=List[ConfigurationSchema],
    status_code=status.HTTP_200_OK,
)
async def read_configurations_by_client_id(
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.CONFIG_READ, commit=False)
    ),
):
    """Return the configurations with the specified client_id."""
    client_id = secure_session.identity_payload.client_id

    if not client_id:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=("Couldn't find a valid client_id in the access token."),
        )

    return await crud_configuration.filter(
        db_session=secure_session.session, filter_expressions=[Configuration.cluster_client_id == client_id]
    )


@router.get(
    "/{configuration_id}",
    response_model=ConfigurationSchema,
    status_code=status.HTTP_200_OK,
)
async def read_configuration(
    configuration_id: int,
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.CONFIG_READ, commit=False)
    ),
):
    """Return a configuration with the associated license severs and features with a given id."""
    return await crud_configuration.read(
        db_session=secure_session.session, id=configuration_id, force_refresh=True
    )


@router.put(
    "/{configuration_id}",
    response_model=ConfigurationSchema,
    status_code=status.HTTP_200_OK,
)
async def update_configuration(
    configuration_id: int,
    configuration_update: ConfigurationCompleteUpdateSchema,
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.CONFIG_UPDATE)),
):
    """
    Update a configuration in the database.

    If there are features in the payload, they'll be updated if they have an id,
    or they will be created if the id is None.

    If there are license servers in the payload, they'll updated if they have an id,
    or they will be created if the id is None.

    All resources related to the configuration that aren't present in the payload
    will be deleted.
    """
    if configuration_update.features is not None:
        await crud_configuration.delete_features(
            db_session=secure_session.session,
            configuration_id=configuration_id,
            payload=configuration_update,
        )

        for feature_update in configuration_update.features:
            if feature_update.id:
                await crud_feature.update(
                    db_session=secure_session.session,
                    id=feature_update.id,
                    obj=FeatureUpdateSchema(**feature_update.dict(exclude={"id"})),
                )
            else:
                feature_obj = {
                    "name": feature_update.name,
                    "product_id": feature_update.product_id,
                    "config_id": configuration_id,
                    "reserved": feature_update.reserved,
                }
                await crud_feature.create(
                    db_session=secure_session.session, obj=FeatureCreateSchema(**feature_obj)
                )

    if configuration_update.license_servers is not None:
        await crud_configuration.delete_license_servers(
            db_session=secure_session.session,
            configuration_id=configuration_id,
            payload=configuration_update,
        )

        for license_server_update in configuration_update.license_servers:
            if license_server_update.id:
                await crud_license_server.update(
                    db_session=secure_session.session,
                    id=license_server_update.id,
                    obj=LicenseServerUpdateSchema(**license_server_update.dict(exclude={"id"})),
                )
            else:
                license_server_obj = {
                    "config_id": configuration_id,
                    "host": license_server_update.host,
                    "port": license_server_update.port,
                }
                await crud_license_server.create(
                    db_session=secure_session.session, obj=LicenseServerCreateSchema(**license_server_obj)
                )

    if any(value for _, value in configuration_update.dict(exclude={"features", "license_servers"}).items()):
        return await crud_configuration.update(
            db_session=secure_session.session,
            id=configuration_id,
            obj=ConfigurationUpdateSchema(
                **configuration_update.dict(exclude={"features", "license_servers"})
            ),
        )
    elif configuration_update.features or configuration_update.license_servers:
        return await crud_configuration.read(
            db_session=secure_session.session, id=configuration_id, force_refresh=True
        )
    else:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Please provide a valid field to update configuration.",
        )


@router.delete(
    "/{configuration_id}",
    status_code=status.HTTP_200_OK,
)
async def delete_configuration(
    configuration_id: int,
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.CONFIG_DELETE)),
):
    """
    Delete a configuration from the database.

    This will also delete the features and license servers associated.
    """
    return await crud_configuration.delete(db_session=secure_session.session, id=configuration_id)



"""Configuration schemas for the License Manager API."""
from typing import List, Optional

from pydantic import ConfigDict, BaseModel, Field, PositiveInt

from lm_api.api.schemas.base import BaseCreateSchema, BaseUpdateSchema
from lm_api.api.schemas.feature import (
    FeatureSchema,
    FeatureWithOptionalIdUpdateSchema,
    FeatureWithoutConfigIdCreateSchema,
)
from lm_api.api.schemas.license_server import (
    LicenseServerSchema,
    LicenseServerWithOptionalIdUpdateSchema,
    LicenseServerWithoutConfigIdCreateSchema,
)
from lm_api.constants import LicenseServerType


class ConfigurationCreateSchema(BaseCreateSchema):
    """
    Represents the configuration for a set of features.
    """

    name: str = Field(
        ..., title="Name of the configuration", max_length=255, description="The name of the configuration."
    )
    cluster_client_id: str = Field(
        ...,
        title="Client ID of the cluster",
        max_length=255,
        description="The client ID of the cluster that will use this configuration.",
    )
    grace_time: int = Field(
        60,
        gt=0,
        le=2**31 - 1,
        title="Grace time",
        description="The grace time in seconds for the license's bookings to be retained.",
    )

    type: LicenseServerType = Field(
        ...,
        title="Type of license server",
        description="The type of license server that provides the license.",
    )


class ConfigurationCompleteCreateSchema(BaseCreateSchema):
    """
    Represents the data to create a complete configuration.

    Includes the features and the license servers.
    """

    name: str = Field(
        ..., title="Name of the configuration", max_length=255, description="The name of the configuration."
    )
    cluster_client_id: str = Field(
        ...,
        title="Client ID of the cluster",
        max_length=255,
        description="The client ID of the cluster that will use this configuration.",
    )
    grace_time: int = Field(
        60,
        gt=0,
        le=2**31 - 1,
        title="Grace time",
        description="The grace time in seconds for the license's bookings to be retained.",
    )
    type: LicenseServerType = Field(
        ...,
        title="Type of license server",
        description="The type of license server that provides the license.",
    )
    features: List[FeatureWithoutConfigIdCreateSchema] = Field(
        ...,
        title="Features of the configuration",
        description="The features of the configuration.",
    )
    license_servers: List[LicenseServerWithoutConfigIdCreateSchema] = Field(
        ...,
        title="License servers of the configuration",
        description="The license servers of the configuration.",
    )


class ConfigurationUpdateSchema(BaseUpdateSchema):
    """
    Represents the data for a configuration update.
    """

    name: Optional[str] = Field(
        None,
        title="Name of the configuration",
        max_length=255,
        description="The name of the configuration.",
    )
    cluster_client_id: Optional[str] = Field(
        None,
        title="Client ID of the cluster",
        max_length=255,
        description="The client ID of the cluster that will use this configuration.",
    )
    grace_time: Optional[int] = Field(
        None,
        gt=0,
        le=2**31 - 1,
        title="Grace time",
        description="The grace time in seconds for the license's bookings to be retained.",
    )
    type: Optional[LicenseServerType] = Field(
        None,
        title="Type of license server",
        description="The type of license server that provides the license.",
    )


class ConfigurationCompleteUpdateSchema(BaseUpdateSchema):
    """
    Represents the data to update a complete configuration.

    Includes the features and the license servers.
    """

    name: Optional[str] = Field(
        None,
        title="Name of the configuration",
        max_length=255,
        description="The name of the configuration.",
    )
    cluster_client_id: Optional[str] = Field(
        None,
        title="Client ID of the cluster",
        max_length=255,
        description="The client ID of the cluster that will use this configuration.",
    )
    grace_time: Optional[int] = Field(
        None,
        gt=0,
        le=2**31 - 1,
        title="Grace time",
        description="The grace time in seconds for the license's bookings to be retained.",
    )
    type: Optional[LicenseServerType] = Field(
        None,
        title="Type of license server",
        description="The type of license server that provides the license.",
    )
    features: Optional[List[FeatureWithOptionalIdUpdateSchema]] = Field(
        None,
        title="Features of the configuration",
        description="The features of the configuration.",
    )
    license_servers: Optional[List[LicenseServerWithOptionalIdUpdateSchema]] = Field(
        None,
        title="License servers of the configuration",
        description="The license servers of the configuration.",
    )


class ConfigurationSchema(BaseModel):
    """
    Represents the configuration for a set of features.
    """

    id: int = Field(..., title="ID of the configuration", description="The ID of the configuration.")
    name: str = Field(
        ..., title="Name of the configuration", max_length=255, description="The name of the configuration."
    )
    cluster_client_id: str = Field(
        ...,
        title="Client ID of the cluster",
        max_length=255,
        description="The client ID of the cluster that will use this configuration.",
    )
    features: List[FeatureSchema] = Field(
        ...,
        title="Features of the configuration",
        description="The features of the configuration.",
    )
    license_servers: List[LicenseServerSchema] = Field(
        ...,
        title="License servers of the configuration",
        description="The license servers of the configuration.",
    )
    grace_time: PositiveInt = Field(
        ...,
        title="Grace time",
        description="The grace time in seconds for the license's bookings to be retained.",
    )
    type: LicenseServerType = Field(
        ...,
        title="Type of license server",
        description="The type of license server that provides the license.",
    )
    model_config = ConfigDict(from_attributes=True)



"""Database model for Features."""
from typing import TYPE_CHECKING, List

from sqlalchemy import Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.schema import CheckConstraint, ForeignKey

from lm_api.api.models.crud_base import CrudBase

if TYPE_CHECKING:
    from lm_api.api.models.booking import Booking
    from lm_api.api.models.configuration import Configuration
    from lm_api.api.models.product import Product
else:
    Booking = "Booking"
    Configuration = "Configuration"
    Product = "Product"


class Feature(CrudBase):
    """
    Represents a feature.
    """

    name = mapped_column(String, nullable=False)
    product_id = mapped_column(Integer, ForeignKey("products.id"), nullable=False)
    config_id = mapped_column(Integer, ForeignKey("configs.id"), nullable=False)
    total = mapped_column(Integer, CheckConstraint("total>=0"), default=0, nullable=False)
    used = mapped_column(Integer, CheckConstraint("used>=0"), default=0, nullable=False)
    reserved = mapped_column(Integer, CheckConstraint("reserved>=0"), nullable=False)

    product = relationship(Product, back_populates="features", lazy="selectin")

    bookings: Mapped[List[Booking]] = relationship(
        Booking,
        back_populates="feature",
        lazy="selectin",
        cascade="all, delete-orphan",
        uselist=True,
    )
    configurations: Mapped[List[Configuration]] = relationship(
        Configuration,
        back_populates="features",
        lazy="selectin",
        uselist=True,
    )

    searchable_fields = [name]
    sortable_fields = [name, product_id, config_id, total, used, reserved]

    def __repr__(self):
        return (
            f"Feature(id={self.id}, "
            f"name={self.name}, "
            f"product_id={self.product_id}, "
            f"config_id={self.config_id}, "
            f"total={self.total}, "
            f"used={self.used}, "
            f"reserved={self.reserved})"
        )



from typing import Optional, Annotated

from pydantic import Field, confloat

from lm_api.constants import LogLevelEnum
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """
    App config.
    """

    DEPLOY_ENV: str = "LOCAL"

    # Sentry settings
    SENTRY_DSN: Optional[str] = None
    SENTRY_TRACES_SAMPLE_RATE: Annotated[float, confloat(gt=0, le=1.0)] = 0.01
    SENTRY_SAMPLE_RATE: Annotated[float, confloat(gt=0.0, le=1.0)] = 0.25
    SENTRY_PROFILING_SAMPLE_RATE: Annotated[float, confloat(gt=0.0, le=1.0)] = 0.01

    # vv should be specified as something like /staging
    # to match where the API is deployed in API Gateway
    ASGI_ROOT_PATH: str = ""

    # Database settings
    DATABASE_HOST: str = "localhost"
    DATABASE_USER: str = "local-user"
    DATABASE_PSWD: str = "local-pswd"
    DATABASE_NAME: str = "local-db"
    DATABASE_PORT: int = 5432

    # Test database settings
    TEST_DATABASE_HOST: str = "localhost"
    TEST_DATABASE_USER: str = "test-db-user"
    TEST_DATABASE_PSWD: str = "test-db-pswd"
    TEST_DATABASE_NAME: str = "test-db-name"
    TEST_DATABASE_PORT: int = 5433

    # Enable multi-tenancy so that the database is determined by the client_id in the auth token
    MULTI_TENANCY_ENABLED: bool = Field(False)

    # log level (everything except sql tracing)
    LOG_LEVEL: LogLevelEnum = LogLevelEnum.INFO

    # log level (sql tracing)
    LOG_LEVEL_SQL: Optional[LogLevelEnum] = None

    # Security Settings. For details, see https://github.com/omnivector-solutions/armsec
    ARMASEC_DOMAIN: str
    ARMASEC_DEBUG: bool = Field(False)
    ARMASEC_ADMIN_DOMAIN: Optional[str] = None
    ARMASEC_ADMIN_MATCH_KEY: Optional[str] = None
    ARMASEC_ADMIN_MATCH_VALUE: Optional[str] = None
    ARMASEC_USE_HTTPS: bool = Field(True)
    model_config = SettingsConfigDict(env_file=".env")


settings = Settings()



"""Schema definitions for the cluster app."""

from lm_api.api.schemas.base import BaseCreateSchema, BaseUpdateSchema
from pydantic import ConfigDict, field_serializer
from pydantic_extra_types.pendulum_dt import DateTime


class ClusterStatusSchema(BaseCreateSchema, BaseUpdateSchema):
    """
    Represents the data for a cluster status to be created.
    """

    cluster_client_id: str
    interval: int
    last_reported: DateTime
    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)

    @field_serializer("last_reported", when_used="json")
    @classmethod
    def serialize_last_reported(cls, value: DateTime) -> str:
        return value.isoformat().replace("Z", "+00:00")



"""
Persistent data storage for the API.
"""
import typing
from dataclasses import dataclass

from fastapi import Depends
from fastapi.exceptions import HTTPException
from loguru import logger
from sqlalchemy import or_
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
from sqlalchemy.orm import Mapped, MappedColumn
from sqlalchemy.sql.expression import ColumnElement, UnaryExpression
from starlette import status
from yarl import URL

from lm_api.config import settings
from lm_api.security import IdentityPayload, PermissionMode, lockdown_with_identity


def build_db_url(
    override_db_name: typing.Optional[str] = None,
    force_test: bool = False,
    asynchronous: bool = True,
) -> str:
    """
    Build a database url based on settings.

    If ``force_test`` is set, build from the test database settings.
    If ``asynchronous`` is set, use asyncpg.
    If ``override_db_name`` replace the database name in the settings with the supplied value.
    """
    prefix = "TEST_" if force_test else ""
    db_user = getattr(settings, f"{prefix}DATABASE_USER")
    db_password = getattr(settings, f"{prefix}DATABASE_PSWD")
    db_host = getattr(settings, f"{prefix}DATABASE_HOST")
    db_port = getattr(settings, f"{prefix}DATABASE_PORT")
    db_name = getattr(settings, f"{prefix}DATABASE_NAME") if override_db_name is None else override_db_name
    db_path = "/{}".format(db_name)
    db_scheme = "postgresql+asyncpg" if asynchronous else "postgresql"

    return str(
        URL.build(
            scheme=db_scheme,
            user=db_user,
            password=db_password,
            host=db_host,
            port=db_port,
            path=db_path,
        )
    )


class EngineFactory:
    """
    Provide a factory class that creates engines and keeps track of them in an engine mapping.

    This is used for multi-tenancy and database URL creation at request time.
    """

    engine_map: typing.Dict[str, AsyncEngine]

    def __init__(self):
        """
        Initialize the EngineFactory.
        """
        self.engine_map = dict()

    async def cleanup(self):
        """
        Close all engines stored in the engine map and clears the engine_map.
        """
        for engine in self.engine_map.values():
            await engine.dispose()
        self.engine_map = dict()

    def get_engine(self, override_db_name: typing.Optional[str] = None) -> AsyncEngine:
        """
        Get a database engine.

        If the database url is already in the engine map, return the engine stored there. Otherwise, build
        a new one, store it, and return the new engine.
        """
        db_url = build_db_url(
            override_db_name=override_db_name,
            force_test=settings.DEPLOY_ENV.lower() == "test",
        )
        if db_url not in self.engine_map:
            self.engine_map[db_url] = create_async_engine(db_url, pool_pre_ping=True)
        return self.engine_map[db_url]

    def get_session(self, override_db_name: typing.Optional[str] = None) -> AsyncSession:
        """
        Get an asynchronous database session.

        Gets a new session from the correct engine in the engine map.
        """
        engine = self.get_engine(override_db_name=override_db_name)
        return AsyncSession(engine, expire_on_commit=False)


engine_factory = EngineFactory()


@dataclass
class SecureSession:
    """
    Provide a container class for an IdentityPayload and AsyncSesson for the current request.
    """

    identity_payload: IdentityPayload
    session: AsyncSession


def secure_session(*scopes: str, permission_mode: PermissionMode = PermissionMode.SOME, commit: bool = True):
    """
    Provide an injectable for FastAPI that checks permissions and returns a database session for this request.

    This should be used for all secured routes that need access to the database. It will commit the
    transaction upon completion of the request. If an exception occurs, it will rollback the transaction.
    If multi-tenancy is enabled, it will retrieve a database session for the database associated with the
    client_id found in the requesting user's auth token.

    If testing mode is enabled, it will flush the session instead of committing changes to the database.

    Note that the session should NEVER be explicitly committed anywhere else in the source code.
    """

    async def dependency(
        identity_payload: IdentityPayload = Depends(
            lockdown_with_identity(*scopes, permission_mode=permission_mode)
        ),
    ) -> typing.AsyncIterator[SecureSession]:
        override_db_name = identity_payload.organization_id if settings.MULTI_TENANCY_ENABLED else None
        session = engine_factory.get_session(override_db_name=override_db_name)
        await session.begin_nested()
        try:
            yield SecureSession(
                identity_payload=identity_payload,
                session=session,
            )
            # In test mode, we should not commit to the database. Instead, just flush to the session
            if settings.DEPLOY_ENV.lower() == "test":
                logger.debug("Flushing session due to test mode")
                await session.flush()
            elif commit is True:
                logger.debug("Committing session")
                await session.commit()
        except Exception as err:
            logger.warning(f"Rolling back session due to error: {err}")
            await session.rollback()
            raise err
        finally:
            # In test mode, we should not close the session so that assertions can be made about the state
            # of the db session in the test functions after calling the application logic
            if settings.DEPLOY_ENV.lower() != "test":
                logger.debug("Closing session")
                await session.close()

    return dependency


def render_sql(query) -> str:
    """
    Render a sqlalchemy query into a string for debugging.
    """
    return query.compile(compile_kwargs={"literal_binds": True})


def search_clause(
    search_terms: str,
    searchable_fields: typing.List[MappedColumn[typing.Any]],
) -> ColumnElement[bool]:
    """
    Create search clause across searchable fields with search terms.
    """
    return or_(*[field.ilike(f"%{term}%") for field in searchable_fields for term in search_terms.split()])


def sort_clause(
    sort_field: str,
    sortable_fields: typing.List[MappedColumn[typing.Any]],
    sort_ascending: bool,
) -> typing.Union[Mapped[typing.Any], UnaryExpression]:
    """
    Create a sort clause given a sort field, the list of sortable fields, and a sort_ascending flag.
    """
    sort_field_names = [f.name for f in sortable_fields]
    try:
        index = sort_field_names.index(sort_field)
    except ValueError:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Invalid sorting column requested: {sort_field}. Must be one of {sort_field_names}",
        )
    sort_column: typing.Union[Mapped[typing.Any], UnaryExpression] = sortable_fields[index]
    if not sort_ascending:
        sort_column = sort_column.desc()
    return sort_column



"""License Server schemas for the License Manager API."""
import re
from typing import Optional

from pydantic import field_validator, ConfigDict, BaseModel, Field, PositiveInt

from lm_api.api.schemas.base import BaseCreateSchema, BaseUpdateSchema


class LicenseServerCreateSchema(BaseCreateSchema):
    """
    License server to be created in the database.
    """

    config_id: int = Field(
        ...,
        title="Configuration ID",
        description="The ID of the configuration that the license server belongs to.",
    )
    host: str = Field(..., title="Host", description="The host of the license server.")
    port: PositiveInt = Field(..., title="Port", description="The port of the license server.")

    @field_validator("host")
    @classmethod
    def validate_host(cls, value):
        HOST = r"^(?!-)[A-Za-z0-9-]{1,63}(?<!-)(\.[A-Za-z0-9-]{1,63})*$"  # noqa: W605
        if not re.match(HOST, value):
            raise ValueError("Host must match the regex [a-z0-9-.]+")
        return value

    @field_validator("port")
    @classmethod
    def validate_port(cls, value):
        if value > 65535:
            raise ValueError("Port must be less than or equal to 65535")
        return value


class LicenseServerWithoutConfigIdCreateSchema(BaseCreateSchema):
    """
    License server to be created by configurations endpoint.

    The config_id will be handled in the endpoint.
    """

    host: str = Field(..., title="Host", description="The host of the license server.")
    port: PositiveInt = Field(..., title="Port", description="The port of the license server.")

    @field_validator("host")
    @classmethod
    def validate_host(cls, value):
        HOST = r"^(?!-)[A-Za-z0-9-]{1,63}(?<!-)(\.[A-Za-z0-9-]{1,63})*$"  # noqa: W605
        if not re.match(HOST, value):
            raise ValueError("Host must match the regex [a-z0-9-.]+")
        return value

    @field_validator("port")
    @classmethod
    def validate_port(cls, value):
        if value > 65535:
            raise ValueError("Port must be less than or equal to 65535")
        return value


class LicenseServerWithOptionalIdUpdateSchema(BaseUpdateSchema):
    """
    License server to be updated in the database.
    """

    id: Optional[int] = Field(None, title="ID", description="The ID of the license server.")
    host: Optional[str] = Field(None, title="Host", description="The host of the license server.")
    port: Optional[PositiveInt] = Field(None, title="Port", description="The port of the license server.")

    @field_validator("host")
    @classmethod
    def validate_host(cls, value):
        HOST = r"^(?!-)[A-Za-z0-9-]{1,63}(?<!-)(\.[A-Za-z0-9-]{1,63})*$"  # noqa: W605
        if not re.match(HOST, value):
            raise ValueError("Host must match the regex [a-z0-9-.]+")
        return value

    @field_validator("port")
    @classmethod
    def validate_port(cls, value):
        if value > 65535:
            raise ValueError("Port must be less than or equal to 65535")
        return value


class LicenseServerUpdateSchema(BaseUpdateSchema):
    """
    License server to be updated in the database.
    """

    config_id: Optional[int] = Field(
        None,
        title="Configuration ID",
        description="The ID of the configuration that the license server belongs to.",
    )
    host: Optional[str] = Field(None, title="Host", description="The host of the license server.")
    port: Optional[PositiveInt] = Field(None, title="Port", description="The port of the license server.")

    @field_validator("host")
    @classmethod
    def validate_host(cls, value):
        HOST = r"^(?!-)[A-Za-z0-9-]{1,63}(?<!-)(\.[A-Za-z0-9-]{1,63})*$"  # noqa: W605
        if not re.match(HOST, value):
            raise ValueError("Host must match the regex [a-z0-9-.]+")
        return value

    @field_validator("port")
    @classmethod
    def validate_port(cls, value):
        if value > 65535:
            raise ValueError("Port must be less than or equal to 65535")
        return value


class LicenseServerSchema(BaseModel):
    """
    License server response from the database.
    """

    id: int = Field(..., title="ID", description="The ID of the license server.")
    config_id: int = Field(
        ...,
        title="Configuration ID",
        description="The ID of the configuration that the license server belongs to.",
    )
    host: str = Field(..., title="Host", description="The host of the license server.")
    port: PositiveInt = Field(..., title="Port", description="The port of the license server.")
    model_config = ConfigDict(from_attributes=True)



from typing import List, Optional

from fastapi import APIRouter, Body, Depends, HTTPException, Query, status

from lm_api.api.cruds.booking import BookingCRUD
from lm_api.api.cruds.feature import FeatureCRUD
from lm_api.api.cruds.generic import GenericCRUD
from lm_api.api.models.booking import Booking
from lm_api.api.models.feature import Feature
from lm_api.api.models.job import Job
from lm_api.api.schemas.booking import BookingCreateSchema
from lm_api.api.schemas.job import JobCreateSchema, JobSchema, JobWithBookingCreateSchema
from lm_api.database import SecureSession, secure_session
from lm_api.permissions import Permissions

router = APIRouter()


crud_job = GenericCRUD(Job)
crud_booking = BookingCRUD(Booking)
crud_feature = FeatureCRUD(Feature)


@router.post(
    "",
    response_model=JobSchema,
    status_code=status.HTTP_201_CREATED,
)
async def create_job(
    job: JobWithBookingCreateSchema = Body(..., description="Job to be created"),
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.JOB_CREATE)),
):
    """Create a new job."""
    client_id = secure_session.identity_payload.client_id

    if not client_id:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=("Couldn't find a valid client_id in the access token."),
        )

    if job.cluster_client_id is None:
        job.cluster_client_id = client_id

    job_created = await crud_job.create(
        db_session=secure_session.session, obj=JobCreateSchema(**job.dict(exclude={"bookings"}))
    )

    if job.bookings:
        try:
            for booking in job.bookings:
                product, feature = booking.product_feature.split(".")
                feature_obj: Feature = await crud_feature.filter_by_product_feature_and_client_id(
                    db_session=secure_session.session,
                    product_name=product,
                    feature_name=feature,
                    client_id=client_id,
                )
                if not feature:
                    raise HTTPException(
                        status_code=status.HTTP_400_BAD_REQUEST,
                        detail=("Couldn't find the feature to book."),
                    )
                obj = {
                    "job_id": job_created.id,
                    "feature_id": feature_obj.id,
                    "quantity": booking.quantity,
                }
                await crud_booking.create(db_session=secure_session.session, obj=BookingCreateSchema(**obj))
        except HTTPException:
            await crud_job.delete(db_session=secure_session.session, id=job_created.id)
            raise

    return await crud_job.read(db_session=secure_session.session, id=job_created.id, force_refresh=True)


@router.get(
    "/by_client_id",
    response_model=List[JobSchema],
    status_code=status.HTTP_200_OK,
)
async def read_jobs_by_client_id(
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.JOB_READ, commit=False)
    ),
):
    """Return the jobs with the specified OIDC client_id retrieved from the request."""
    client_id = secure_session.identity_payload.client_id

    if not client_id:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=("Couldn't find a valid client_id in the access token."),
        )

    return await crud_job.filter(
        db_session=secure_session.session, filter_expressions=[Job.cluster_client_id == client_id]
    )


@router.get(
    "",
    response_model=List[JobSchema],
    status_code=status.HTTP_200_OK,
)
async def read_all_jobs(
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.JOB_READ, commit=False)
    ),
):
    """Return all jobs."""
    return await crud_job.read_all(
        db_session=secure_session.session,
        search=search,
        sort_field=sort_field,
        sort_ascending=sort_ascending,
    )


@router.get(
    "/{job_id}",
    response_model=JobSchema,
    status_code=status.HTTP_200_OK,
)
async def read_job(
    job_id: int,
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.JOB_READ, commit=False)
    ),
):
    """Return a job with associated bookings with the given id."""
    return await crud_job.read(db_session=secure_session.session, id=job_id, force_refresh=True)


@router.delete(
    "/{job_id}",
    status_code=status.HTTP_200_OK,
)
async def delete_job(
    job_id: int,
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.JOB_DELETE)),
):
    """Delete a job from the database and associated bookings."""
    return await crud_job.delete(db_session=secure_session.session, id=job_id)


@router.delete(
    "/slurm_job_id/{slurm_job_id}",
    status_code=status.HTTP_200_OK,
)
async def delete_job_by_slurm_id(
    slurm_job_id: str,
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.JOB_DELETE)),
):
    """
    Delete a job from the database and associated bookings.

    Uses the slurm_job_id and the cluster client_id to filter the job.

    Since the slurm_job_id can be the same across clusters, we need the cluster client_id to validate.
    """
    client_id = secure_session.identity_payload.client_id

    if not client_id:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=("Couldn't find a valid client_id in the access token."),
        )

    jobs = await crud_job.filter(
        db_session=secure_session.session,
        filter_expressions=[Job.slurm_job_id == slurm_job_id, Job.cluster_client_id == client_id],
    )

    for job in jobs:
        return await crud_job.delete(db_session=secure_session.session, id=job.id)

    raise HTTPException(status_code=404, detail="The job doesn't exist in this cluster.")


@router.get(
    "/slurm_job_id/{slurm_job_id}",
    response_model=JobSchema,
    status_code=status.HTTP_200_OK,
)
async def read_job_by_slurm_id(
    slurm_job_id: str,
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.JOB_READ, commit=False)
    ),
):
    """
    Read a job from the database and associated bookings.

    Uses the slurm_job_id and the cluster client_id to filter the job.

    Since the slurm_job_id can be the same across clusters, we need the cluster client_id to validate.
    """
    client_id = secure_session.identity_payload.client_id

    if not client_id:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=("Couldn't find a valid client_id in the access token."),
        )

    jobs = await crud_job.filter(
        db_session=secure_session.session,
        filter_expressions=[Job.slurm_job_id == slurm_job_id, Job.cluster_client_id == client_id],
    )

    for job in jobs:
        return job

    raise HTTPException(status_code=404, detail="The job doesn't exist in this cluster.")



"""Cluster status API endpoints."""
from typing import List, Optional

from fastapi import APIRouter, Depends, Query, HTTPException, status
from pendulum.datetime import DateTime as PendulumDateTime

from lm_api.api.cruds.cluster_status import ClusterStatusCRUD
from lm_api.api.models.cluster_status import ClusterStatus
from lm_api.api.schemas.cluster_status import ClusterStatusSchema
from lm_api.database import SecureSession, secure_session
from lm_api.permissions import Permissions
from buzz import require_condition


router = APIRouter()
crud_cluster_status = ClusterStatusCRUD(ClusterStatus)


@router.put(
    "",
    status_code=status.HTTP_202_ACCEPTED,
    description="Endpoints to accept a status check from the agent on the clusters.",
)
async def report_cluster_status(
    interval: int = Query(description="The interval in seconds between pings.", gt=0),
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.STATUS_UPDATE)),
):
    """
    Report the status of the cluster.
    """
    client_id = secure_session.identity_payload.client_id

    if not client_id:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=("Couldn't find a valid client_id in the access token."),
        )

    cluster_status = ClusterStatusSchema(
        cluster_client_id=secure_session.identity_payload.client_id,
        interval=interval,
        last_reported=PendulumDateTime.utcnow(),
    )

    return await crud_cluster_status.upsert(db_session=secure_session.session, payload=cluster_status)


@router.get(
    "",
    description="Endpoint to get the status of all clusters.",
    response_model=List[ClusterStatusSchema],
)
async def read_all_cluster_statuses(
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.STATUS_READ, commit=False)
    ),
):
    """
    Get the status of the cluster.
    """
    return await crud_cluster_status.read_all(
        db_session=secure_session.session,
    )


@router.get(
    "/{cluster_client_id}",
    description="Endpoint to get the status of a specific cluster.",
    response_model=Optional[ClusterStatusSchema],
)
async def read_cluster_status_by_client_id(
    cluster_client_id: str,
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.STATUS_READ, commit=False)
    ),
):
    """
    Get the status of a specific cluster.

    There should only be one status per cluster.
    """
    cluster_statuses = await crud_cluster_status.filter(
        db_session=secure_session.session,
        filter_expressions=[ClusterStatus.cluster_client_id == cluster_client_id],
    )

    require_condition(
        len(cluster_statuses) <= 1,
        "There should only be one status per cluster.",
        raise_exc_class=HTTPException,
        raise_kwargs=dict(
            status_code=status.HTTP_400_BAD_REQUEST, detail="There should only be one status per cluster."
        ),
    )

    if not cluster_statuses:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Cluster with client_id {cluster_client_id} not found.",
        )

    return cluster_statuses[0]



"""Feature CRUD class for SQLAlchemy models."""
from fastapi import HTTPException
from loguru import logger
from sqlalchemy import delete
from sqlalchemy.ext.asyncio import AsyncSession

from lm_api.api.cruds.generic import GenericCRUD
from lm_api.api.models.feature import Feature
from lm_api.api.models.license_server import LicenseServer
from lm_api.api.schemas.configuration import ConfigurationCompleteUpdateSchema


class ConfigurationCRUD(GenericCRUD):
    """Configuration CRUD module to implement configuration update."""

    async def delete_features(
        self, db_session: AsyncSession, configuration_id: int, payload: ConfigurationCompleteUpdateSchema
    ):
        """
        Delete all features that aren't in the payload.
        """
        features_in_payload = [f.id for f in payload.features if f.id is not None] if payload.features else []
        delete_unused_features_query = (
            delete(Feature)
            .where(~Feature.id.in_(features_in_payload))
            .where(Feature.config_id == configuration_id)
        )

        try:
            await db_session.execute(delete_unused_features_query)
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail="Feature could not be deleted.")

    async def delete_license_servers(
        self, db_session: AsyncSession, configuration_id: int, payload: ConfigurationCompleteUpdateSchema
    ):
        """
        Delete all license servers that aren't in the payload.
        """
        license_servers_in_payload = (
            [ls.id for ls in payload.license_servers if ls.id is not None] if payload.license_servers else []
        )
        delete_unused_license_servers_query = (
            delete(LicenseServer)
            .where(~LicenseServer.id.in_(license_servers_in_payload))
            .where(LicenseServer.config_id == configuration_id)
        )

        try:
            await db_session.execute(delete_unused_license_servers_query)
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail="License Server could not be deleted.")



"""Job schemas for the License Manager API."""
from typing import List, Optional

from pydantic import ConfigDict, BaseModel, Field

from lm_api.api.schemas.base import BaseCreateSchema, BaseUpdateSchema
from lm_api.api.schemas.booking import BookingSchema


class JobBookingCreateSchema(BaseCreateSchema):
    product_feature: str = Field(
        ..., title="Product feature", description="The product.feature of the license."
    )
    quantity: int = Field(
        ..., gt=0, le=2**31 - 1, title="Quantity", description="The quantity of booked licenses."
    )


class JobCreateSchema(BaseCreateSchema):
    """
    Represents the jobs submitted in a cluster.
    """

    slurm_job_id: str = Field(..., title="Slurm job ID", description="The ID of the job in the cluster.")
    cluster_client_id: str = Field(
        ...,
        title="Client ID of the cluster",
        description="The client ID of the cluster that submitted the job.",
    )
    username: str = Field(
        ..., title="Username", description="The username of the user that submitted the job."
    )
    lead_host: str = Field(
        ..., title="Lead host", description="The lead host of the cluster that submitted the job."
    )


class JobWithBookingCreateSchema(BaseCreateSchema):
    """
    Represents the jobs submitted in a cluster with its bookings.
    """

    slurm_job_id: str = Field(..., title="Slurm job ID", description="The ID of the job in the cluster.")
    cluster_client_id: Optional[str] = Field(
        None,
        title="Client ID of the cluster",
        description="The client ID of the cluster that submitted the job.",
    )
    username: str = Field(
        ..., title="Username", description="The username of the user that submitted the job."
    )
    lead_host: str = Field(
        ..., title="Lead host", description="The lead host of the cluster that submitted the job."
    )
    bookings: List[JobBookingCreateSchema] = Field(
        ..., title="Bookings", description="The bookings of the job."
    )


class JobUpdateSchema(BaseUpdateSchema):
    """
    Represents the jobs submitted in a cluster.
    """

    slurm_job_id: Optional[str] = Field(
        None, title="Slurm job ID", description="The ID of the job in the cluster."
    )
    cluster_client_id: Optional[str] = Field(
        None,
        title="Client ID of the cluster",
        description="The client ID of the cluster that submitted the job.",
    )
    username: Optional[str] = Field(
        None, title="Username", description="The username of the user that submitted the job."
    )
    lead_host: Optional[str] = Field(
        None, title="Lead host", description="The lead host of the cluster that submitted the job."
    )


class JobSchema(BaseModel):
    """
    Represents the jobs submitted in a cluster.
    """

    id: int = Field(..., title="ID", description="The ID of the job.")
    slurm_job_id: str = Field(..., title="Slurm job ID", description="The ID of the job in the cluster.")
    cluster_client_id: str = Field(
        ...,
        title="Client ID of the cluster",
        description="The client ID of the cluster that submitted the job.",
    )
    username: str = Field(
        ..., title="Username", description="The username of the user that submitted the job."
    )
    lead_host: str = Field(
        ..., title="Lead host", description="The lead host of the cluster that submitted the job."
    )

    bookings: Optional[List[BookingSchema]] = Field(
        None, title="Bookings", description="The bookings of the job."
    )
    model_config = ConfigDict(from_attributes=True)



"""Database model for License Servers."""

from typing import TYPE_CHECKING, List

from sqlalchemy import Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.schema import CheckConstraint, ForeignKey

from lm_api.api.models.crud_base import CrudBase

if TYPE_CHECKING:
    from lm_api.api.models.configuration import Configuration
else:
    Configuration = "Configuration"


class LicenseServer(CrudBase):
    """
    Represents the license servers in a feature configuration.
    """

    config_id = mapped_column(Integer, ForeignKey("configs.id"), nullable=False)
    host = mapped_column(String, nullable=False)
    port = mapped_column(Integer, CheckConstraint("port>0"), nullable=False)

    configurations: Mapped[List[Configuration]] = relationship(
        Configuration,
        back_populates="license_servers",
        lazy="selectin",
    )

    searchable_fields = [host]
    sortable_fields = [config_id, host]

    def __repr__(self):
        return f"LicenseServer(id={self.id}, host={self.host}, port={self.port})"



"""Database models for the cluster health resource."""

from pendulum.datetime import DateTime as PendulumDateTime
from sqlalchemy import DateTime, Integer, String
from lm_api.api.models.crud_base import CrudWithoutId

from inflection import tableize
from sqlalchemy.orm import Mapped, declared_attr, mapped_column


class ClusterStatus(CrudWithoutId):
    """Cluster status table definition."""

    @declared_attr.directive
    @classmethod
    def __tablename__(cls) -> str:
        """
        Dynamically create table name based on the class name.
        """
        return tableize(cls.__name__)

    cluster_client_id: Mapped[str] = mapped_column(String, primary_key=True)
    interval: Mapped[int] = mapped_column(Integer, nullable=False)
    last_reported: Mapped[PendulumDateTime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        default=PendulumDateTime.utcnow,
    )

    @property
    def is_healthy(self) -> bool:
        """Return True if the last_reported time is before now plus the interval in seconds between pings."""
        return PendulumDateTime.utcnow().subtract(seconds=self.interval) <= self.last_reported



from typing import Any, List

from inflection import tableize
from sqlalchemy import Integer
from sqlalchemy.orm import DeclarativeBase, MappedColumn, declared_attr, mapped_column


class Base(DeclarativeBase):
    pass


class CrudBase(Base):
    __abstract__ = True  # Mark this class as abstract to avoid creating a table

    @declared_attr.directive
    @classmethod
    def __tablename__(cls) -> str:
        """
        Dynamically create table name based on the class name.
        """
        return tableize(cls.__name__)

    id = mapped_column(Integer, primary_key=True)

    sortable_fields: List[MappedColumn[Any]] = []
    searchable_fields: List[MappedColumn[Any]] = []


class CrudWithoutId(CrudBase):
    __abstract__ = True

    @declared_attr.directive
    @classmethod
    def id(cls):
        return None



from enum import Enum


class Permissions(str, Enum):
    """
    Describe the permissions that may be used for protecting LM Backend routes.
    """

    ADMIN = "license-manager:admin"

    STATUS_UPDATE = "license-manager:status:update"
    STATUS_READ = "license-manager:status:read"

    LICENSE_SERVER_CREATE = "license-manager:license-server:create"
    LICENSE_SERVER_READ = "license-manager:license-server:read"
    LICENSE_SERVER_UPDATE = "license-manager:license-server:update"
    LICENSE_SERVER_DELETE = "license-manager:license-server:delete"

    CONFIG_CREATE = "license-manager:config:create"
    CONFIG_READ = "license-manager:config:read"
    CONFIG_UPDATE = "license-manager:config:update"
    CONFIG_DELETE = "license-manager:config:delete"

    PRODUCT_CREATE = "license-manager:product:create"
    PRODUCT_READ = "license-manager:product:read"
    PRODUCT_UPDATE = "license-manager:product:update"
    PRODUCT_DELETE = "license-manager:product:delete"

    FEATURE_CREATE = "license-manager:feature:create"
    FEATURE_READ = "license-manager:feature:read"
    FEATURE_UPDATE = "license-manager:feature:update"
    FEATURE_DELETE = "license-manager:feature:delete"

    JOB_CREATE = "license-manager:job:create"
    JOB_READ = "license-manager:job:read"
    JOB_UPDATE = "license-manager:job:update"
    JOB_DELETE = "license-manager:job:delete"

    BOOKING_CREATE = "license-manager:booking:create"
    BOOKING_READ = "license-manager:booking:read"
    BOOKING_UPDATE = "license-manager:booking:update"
    BOOKING_DELETE = "license-manager:booking:delete"



"""
Instantiates armada-security resources for auth on api endpoints using project settings.
Also provides a factory function for TokenSecurity to reduce boilerplate.
"""

import typing

from typing_extensions import Self

from armasec import Armasec, TokenPayload
from armasec.token_security import PermissionMode
from fastapi import Depends
from loguru import logger
from pydantic import model_validator, EmailStr

from lm_api.config import settings


guard = Armasec(
    domain=settings.ARMASEC_DOMAIN,
    debug_logger=logger.debug if settings.ARMASEC_DEBUG else None,
    use_https=settings.ARMASEC_USE_HTTPS,
)


class IdentityPayload(TokenPayload):
    """
    Provide an extension of TokenPayload that includes the user's identity.
    """

    email: typing.Optional[EmailStr] = None
    organization: typing.Optional[typing.Union[str, typing.Dict]] = None
    organization_id: typing.Optional[str] = None

    @model_validator(mode="after")
    def extract_organization(self) -> Self:
        """
        Extracts the organization_id from the organization payload.

        The payload is expected to look like:
        {
            ...,
            "organization": {
                "adf99e01-5cd5-41ac-a1af-191381ad7780": {
                    ...
                }
            }
        }
        """
        if self.organization is None:
            return typing.cast(Self, self)
        elif len(self.organization) != 1 and isinstance(self.organization, dict):
            raise ValueError(
                f"Organization payload did not include exactly one value: {getattr(self, 'organization')}"
            )
        setattr(self, "organization_id", next(iter(self.organization)))
        return typing.cast(Self, self)


def lockdown_with_identity(*scopes: str, permission_mode: PermissionMode = PermissionMode.SOME):
    """
    Provide a wrapper to be used with dependency injection to extract identity on a secured route.
    """

    def dependency(
        token_payload: TokenPayload = Depends(guard.lockdown(*scopes, permission_mode=permission_mode)),
    ) -> IdentityPayload:
        """
        Provide an injectable function to lockdown a route and extract the identity payload.
        """
        return IdentityPayload.model_validate(token_payload, from_attributes=True)

    return dependency



"""Booking schemas for the License Manager API."""
from typing import Optional

from pydantic import ConfigDict, BaseModel, Field, PositiveInt

from lm_api.api.schemas.base import BaseCreateSchema, BaseUpdateSchema


class BookingCreateSchema(BaseCreateSchema):
    """
    Represents the booking of a feature.
    """

    job_id: int = Field(..., title="Job ID", description="The ID of the job that booked the feature.")
    feature_id: int = Field(..., title="Feature ID", description="The ID of the feature that was booked.")
    quantity: int = Field(
        ...,
        gt=0,
        le=2**31 - 1,
        title="Quantity",
        description="The quantity of the feature that was booked.",
    )


class BookingUpdateSchema(BaseUpdateSchema):
    """
    Represents the booking of a feature.
    """

    job_id: Optional[int] = Field(
        None, title="Job ID", description="The ID of the job that booked the feature."
    )
    feature_id: Optional[int] = Field(
        None, title="Feature ID", description="The ID of the feature that was booked."
    )
    quantity: Optional[int] = Field(
        None,
        gt=0,
        le=2**31 - 1,
        title="Quantity",
        description="The quantity of the feature that was booked.",
    )


class BookingSchema(BaseModel):
    """
    Represents the booking of a feature.
    """

    id: int = Field(..., title="ID", description="The ID of the booking.")
    job_id: int = Field(..., title="Job ID", description="The ID of the job that booked the feature.")
    feature_id: int = Field(..., title="Feature ID", description="The ID of the feature that was booked.")
    quantity: PositiveInt = Field(
        ..., title="Quantity", description="The quantity of the feature that was booked."
    )
    model_config = ConfigDict(from_attributes=True)



"""Feature CRUD class for SQLAlchemy models."""
from typing import List, Optional, Sequence, Union

from fastapi import HTTPException
from loguru import logger
from sqlalchemy import Column, func, select, tuple_
from sqlalchemy.ext.asyncio import AsyncSession

from lm_api.api.cruds.generic import GenericCRUD
from lm_api.api.models.booking import Booking
from lm_api.api.models.configuration import Configuration
from lm_api.api.models.feature import Feature
from lm_api.api.models.product import Product
from lm_api.api.schemas.feature import FeatureSchema, FeatureUpdateByNameSchema
from lm_api.database import search_clause, sort_clause


class FeatureCRUD(GenericCRUD):
    """Feature CRUD module to implement feature bulk update and lookup."""

    async def read(
        self, db_session: AsyncSession, id: Union[Column[int], int], force_refresh: bool = False
    ) -> Optional[Feature]:
        """
        Read a Feature object from the database with the given id.
        Returns the object or raise an exception if it does not exist.
        """
        stmt = (
            select(
                *(Feature.__table__.c),
                Product.id.label("product_id"),
                Product.name.label("product_name"),
                func.coalesce(func.sum(Booking.quantity), 0).label("booked_total"),
            )
            .join(Product, Feature.product_id == Product.id)
            .join(Booking, Feature.id == Booking.feature_id, isouter=True)
            .where(Feature.id == id)
            .group_by(Feature.id, Product.id)
        )

        try:
            query = await db_session.execute(stmt)
            db_obj = query.first()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be read.")

        if db_obj is None:
            raise HTTPException(status_code=404, detail=f"{self.model.__name__} not found.")

        return FeatureSchema.from_flat_dict(db_obj._asdict())

    async def bulk_update(
        self, db_session: AsyncSession, features: List[FeatureUpdateByNameSchema], cluster_client_id: str
    ):
        """
        Update a list of features in the database. Since features with the same name can exist in different
        clusters, the client_id is used to filter the cluster where the feature is located.
        """
        get_features_query = (
            select(Feature)
            .join(Product, Feature.product_id == Product.id)
            .join(Configuration, Feature.config_id == Configuration.id)
            .where(
                tuple_(
                    Product.name,
                    Feature.name,
                    Configuration.cluster_client_id,
                ).in_(
                    [(feature.product_name, feature.feature_name, cluster_client_id) for feature in features]
                )
            )
        )

        try:
            result = await db_session.execute(get_features_query)
            db_objs: Sequence[Feature] = result.scalars().all()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail="Feature could not be read.")

        if len(db_objs) != len(features):
            raise HTTPException(status_code=404, detail="Feature not found.")

        for db_obj in db_objs:
            for feature in features:
                if db_obj.product.name == feature.product_name and db_obj.name == feature.feature_name:
                    db_obj.total = feature.total
                    db_obj.used = feature.used
                    break

        try:
            await db_session.flush()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail="Feature could not be updated.")

    async def filter_by_product_feature_and_client_id(
        self, db_session: AsyncSession, product_name: str, feature_name: str, client_id: str
    ) -> Feature:
        """
        Filter features using the product name and feature name as a filter.
        Since the name is not unique across clusters, the client_id
        in the token is used to identify the cluster.
        """
        filter_features_query = (
            select(Feature)
            .join(Product, Feature.product_id == Product.id)
            .join(Configuration, Feature.config_id == Configuration.id)
            .where(
                tuple_(
                    Product.name,
                    Feature.name,
                    Configuration.cluster_client_id,
                ).in_([(product_name, feature_name, client_id)])
            )
        )
        try:
            query = await db_session.execute(filter_features_query)
            db_obj = query.scalars().one_or_none()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail="Feature could not be read.")

        if db_obj is None:
            raise HTTPException(status_code=404, detail="Feature not found.")

        return db_obj

    async def read_all(
        self,
        db_session: AsyncSession,
        search: Optional[str] = None,
        sort_field: Optional[str] = None,
        sort_ascending: bool = True,
    ) -> List[FeatureSchema]:
        """
        Read all objects.
        Returns a list of objects.
        """
        try:
            stmt = (
                select(
                    *(Feature.__table__.c),
                    Product.id.label("product_id"),
                    Product.name.label("product_name"),
                    func.coalesce(func.sum(Booking.quantity), 0).label("booked_total"),
                )
                .join(Product, Feature.product_id == Product.id)
                .join(Booking, Feature.id == Booking.feature_id, isouter=True)
            )
            if search is not None:
                stmt = stmt.where(search_clause(search, self.model.searchable_fields))
            stmt = stmt.group_by(Feature.id, Product.id)
            if sort_field is not None:
                stmt = stmt.order_by(sort_clause(sort_field, self.model.sortable_fields, sort_ascending))
            query = await db_session.execute(stmt)
            return [FeatureSchema.from_flat_dict(r._asdict()) for r in query.all()]
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__}s could not be read.")



"""Product schemas for the License Manager API."""
from typing import Optional

from pydantic import ConfigDict, BaseModel, Field

from lm_api.api.schemas.base import BaseCreateSchema, BaseUpdateSchema


class ProductCreateSchema(BaseCreateSchema):
    """
    Represents a feature's product.
    """

    name: str = Field(
        ..., title="Name of the product", max_length=255, description="The name of the product."
    )


class ProductUpdateSchema(BaseUpdateSchema):
    """
    Represents a feature's product.
    """

    name: Optional[str] = Field(
        None, title="Name of the product", max_length=255, description="The name of the product."
    )


class ProductSchema(BaseModel):
    """
    Represents a feature's product.
    """

    id: int = Field(..., title="ID", description="The ID of the product.")
    name: str = Field(
        ..., title="Name of the product", max_length=255, description="The name of the product."
    )
    model_config = ConfigDict(from_attributes=True)



"""Database model for Bookings."""
from typing import TYPE_CHECKING

from sqlalchemy import Integer, func
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.schema import CheckConstraint, ForeignKey
from sqlalchemy.sql.sqltypes import DateTime

from lm_api.api.models.crud_base import CrudBase

if TYPE_CHECKING:
    from lm_api.api.models.feature import Feature
    from lm_api.api.models.job import Job
else:
    Feature = "Feature"
    Job = "Job"


class Booking(CrudBase):
    """
    Represents the bookings of a feature.
    """

    job_id = mapped_column(Integer, ForeignKey("jobs.id"), nullable=False)
    feature_id = mapped_column(Integer, ForeignKey("features.id"), nullable=False)
    quantity = mapped_column(Integer, CheckConstraint("quantity>=0"), nullable=False)
    created_at = mapped_column(DateTime, default=func.now())

    job: Mapped[Job] = relationship(Job, back_populates="bookings", lazy="selectin")
    feature: Mapped[Feature] = relationship(Feature, back_populates="bookings", lazy="selectin")

    sortable_fields = [job_id, feature_id]

    def __repr__(self):
        return (
            f"Booking(id={self.id}, "
            f"job_id={self.job_id}, "
            f"feature_id={self.feature_id}, "
            f"quantity={self.quantity}, "
            f"created_at={self.created_at})"
        )



"""Database model for Products."""

from typing import TYPE_CHECKING, List

from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from lm_api.api.models.crud_base import CrudBase

if TYPE_CHECKING:
    from lm_api.api.models.feature import Feature
else:
    Feature = "Feature"


class Product(CrudBase):
    """
    Represents a feature's product.
    """

    name = mapped_column(String, nullable=False, unique=True)

    features: Mapped[List[Feature]] = relationship(
        Feature,
        back_populates="product",
        lazy="selectin",
        cascade="all, delete-orphan",
    )

    searchable_fields = [name]
    sortable_fields = [name]

    def __repr__(self):
        return f"Product(id={self.id}, name={self.name})"



"""Cluster CRUD class for SQLAlchemy models."""
from fastapi import HTTPException
from loguru import logger
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from typing import Optional
from lm_api.api.cruds.generic import GenericCRUD
from lm_api.api.models.cluster_status import ClusterStatus
from lm_api.api.schemas.cluster_status import ClusterStatusSchema


class ClusterStatusCRUD(GenericCRUD):
    """Cluster CRUD module to implement PUT endpoint for cluster status update."""

    async def upsert(
        self,
        db_session: AsyncSession,
        payload: ClusterStatusSchema,
    ) -> Optional[ClusterStatus]:
        """Creates a new cluster status update or updates the existing one."""
        try:
            stmt = select(ClusterStatus).filter(ClusterStatus.cluster_client_id == payload.cluster_client_id)
            query = await db_session.execute(stmt)
            db_obj = query.scalar_one_or_none()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be read.")

        if db_obj is None:
            create_obj = self.model(**payload.model_dump())
            try:
                db_session.add(create_obj)
            except Exception as e:
                logger.error(e)
                raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be created.")

            await db_session.flush()
            await db_session.refresh(create_obj)
            return db_obj
        else:
            for field, value in payload:
                if hasattr(db_obj, field) and value is not None:
                    setattr(db_obj, field, value)

            try:
                await db_session.flush()
                await db_session.refresh(db_obj)
            except Exception as e:
                logger.error(e)
                raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be updated.")

            return db_obj



from enum import Enum


class LogLevelEnum(str, Enum):
    """
    Describe the available logging levels.
    """

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class LicenseServerType(str, Enum):
    """
    Describe the supported license server types that may be used for fetching licenses from license servers.
    """

    FLEXLM = "flexlm"
    RLM = "rlm"
    LMX = "lmx"
    LSDYNA = "lsdyna"
    OLICENSE = "olicense"
    DSLS = "dsls"



from typing import List, Optional

from fastapi import APIRouter, Body, Depends, HTTPException, Query, status

from lm_api.api.cruds.feature import FeatureCRUD
from lm_api.api.models.feature import Feature
from lm_api.api.schemas.feature import (
    FeatureCreateSchema,
    FeatureSchema,
    FeatureUpdateByNameSchema,
    FeatureUpdateSchema,
)
from lm_api.database import SecureSession, secure_session
from lm_api.permissions import Permissions

router = APIRouter()

crud_feature = FeatureCRUD(Feature)


@router.post(
    "",
    response_model=FeatureSchema,
    status_code=status.HTTP_201_CREATED,
)
async def create_feature(
    feature: FeatureCreateSchema = Body(..., description="Feature to be created"),
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.FEATURE_CREATE)),
):
    """Create a new feature"""
    return await crud_feature.create(db_session=secure_session.session, obj=feature)


@router.get(
    "",
    response_model=List[FeatureSchema],
    status_code=status.HTTP_200_OK,
)
async def read_all_features(
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.FEATURE_READ, commit=False)
    ),
):
    """Return all features with associated bookings."""
    return await crud_feature.read_all(
        db_session=secure_session.session,
        search=search,
        sort_field=sort_field,
        sort_ascending=sort_ascending,
    )


@router.get(
    "/{feature_id}",
    response_model=FeatureSchema,
    status_code=status.HTTP_200_OK,
)
async def read_feature(
    feature_id: int,
    secure_session: SecureSession = Depends(
        secure_session(Permissions.ADMIN, Permissions.FEATURE_READ, commit=False)
    ),
):
    """Return a feature with associated bookings with the given id."""
    return await crud_feature.read(db_session=secure_session.session, id=feature_id)


@router.put(
    "/bulk",
    status_code=status.HTTP_200_OK,
)
async def bulk_update_feature(
    features: List[FeatureUpdateByNameSchema],
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.FEATURE_UPDATE)),
):
    """
    Update a list of features in the database using the name of each feature.
    Since the name is not unique across clusters, the client_id
    in the token is used to identify the cluster.
    """
    client_id = secure_session.identity_payload.client_id

    if not client_id:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=("Couldn't find a valid client_id in the access token."),
        )

    await crud_feature.bulk_update(
        db_session=secure_session.session,
        features=features,
        cluster_client_id=client_id,
    )

    return {"message": "Features updated successfully."}


@router.put(
    "/{feature_id}",
    response_model=FeatureSchema,
    status_code=status.HTTP_200_OK,
)
async def update_feature(
    feature_id: int,
    feature_update: FeatureUpdateSchema,
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.FEATURE_UPDATE)),
):
    """Update a feature in the database."""
    return await crud_feature.update(
        db_session=secure_session.session,
        id=feature_id,
        obj=feature_update,
    )


@router.delete(
    "/{feature_id}",
    status_code=status.HTTP_200_OK,
)
async def delete_feature(
    feature_id: int,
    secure_session: SecureSession = Depends(secure_session(Permissions.ADMIN, Permissions.FEATURE_DELETE)),
):
    """
    Delete a feature from the database.
    """
    return await crud_feature.delete(db_session=secure_session.session, id=feature_id)



"""Generic CRUD class for SQLAlchemy models."""
from __future__ import annotations

from typing import List, Optional, Sequence, Type, Union

from fastapi import HTTPException
from loguru import logger
from pydantic import BaseModel
from sqlalchemy import Column, ColumnElement, and_, select
from sqlalchemy.ext.asyncio import AsyncSession

from lm_api.api.models.crud_base import CrudBase
from lm_api.api.schemas.base import BaseCreateSchema, BaseUpdateSchema
from lm_api.database import search_clause, sort_clause


class GenericCRUD:
    """Generic CRUD module to interface with database, to be utilized by all models."""

    def __init__(self, model: Type[CrudBase]):
        """Initializes the CRUD class with the model to be used."""
        self.model = model

    async def create(self, db_session: AsyncSession, obj: BaseCreateSchema) -> CrudBase:
        """Creates a new object in the database."""
        db_obj = self.model(**obj.model_dump())
        try:
            db_session.add(db_obj)
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be created.")

        # TODO: Determine if the session actually needs to be flushed here. I think it might not
        await db_session.flush()
        await db_session.refresh(db_obj)
        return db_obj

    async def filter(
        self, db_session: AsyncSession, filter_expressions: List[ColumnElement[bool]]
    ) -> List[CrudBase]:
        """
        Filter objects using a filter field and filter term.
        Returns the list of objects or raise an exception if it does not exist.
        """
        try:
            query = await db_session.execute(select(self.model).filter(and_(*filter_expressions)))
            db_objs = list(query.scalars().all())
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be read.")

        if db_objs is None:
            raise HTTPException(status_code=404, detail=f"{self.model.__name__} not found.")

        return db_objs

    async def read(
        self, db_session: AsyncSession, id: Union[Column[int], int], force_refresh: bool = False
    ) -> Optional[CrudBase]:
        """
        Read an object from the database with the given id.
        Returns the object or raise an exception if it does not exist.
        """
        try:
            query = await db_session.execute(select(self.model).filter(self.model.id == id))
            db_obj = query.scalars().one_or_none()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be read.")

        if db_obj is None:
            raise HTTPException(status_code=404, detail=f"{self.model.__name__} not found.")

        if force_refresh:
            await db_session.refresh(db_obj)

        return db_obj

    async def read_all(
        self,
        db_session: AsyncSession,
        search: Optional[str] = None,
        sort_field: Optional[str] = None,
        sort_ascending: bool = True,
    ) -> Sequence[Union[CrudBase, BaseModel]]:
        """
        Read all objects.
        Returns a list of objects.
        """
        try:
            stmt = select(self.model)
            if search is not None:
                stmt = stmt.where(search_clause(search, self.model.searchable_fields))
            if sort_field is not None:
                stmt = stmt.order_by(sort_clause(sort_field, self.model.sortable_fields, sort_ascending))
            query = await db_session.scalars(stmt)
            return query.all()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__}s could not be read.")

    async def update(
        self,
        db_session: AsyncSession,
        id: Union[Column[int], int],
        obj: BaseUpdateSchema,
    ) -> Optional[CrudBase]:
        """
        Update an object in the database.
        Returns the updated object.
        """
        try:
            query = await db_session.execute(select(self.model).filter(self.model.id == id))
            db_obj = query.scalar_one_or_none()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be updated.")

        if db_obj is None:
            raise HTTPException(status_code=404, detail=f"{self.model.__name__} not found.")

        if all(value is None for _, value in obj):
            raise HTTPException(
                status_code=400, detail=f"Please provide a valid field to update {self.model.__name__}."
            )

        for field, value in obj:
            if hasattr(db_obj, field) and value is not None:
                setattr(db_obj, field, value)

        try:
            await db_session.flush()
            await db_session.refresh(db_obj)
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be updated.")

        return db_obj

    async def delete(self, db_session: AsyncSession, id: Union[Column[int], int]):
        """
        Delete an object from the database.
        """
        try:
            query = await db_session.execute(select(self.model).filter(self.model.id == id))
            db_obj = query.scalar_one_or_none()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be deleted.")

        if db_obj is None:
            raise HTTPException(status_code=404, detail=f"{self.model.__name__} not found.")

        try:
            await db_session.delete(db_obj)
            await db_session.flush()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be deleted.")

        return {"message": f"{self.model.__name__} deleted successfully."}
